@@ -1,15 +1,15 @@
-import pdb
 import abc
-import time
-import os
-import random
-import fnmatch
-import multiprocessing
-import Queue
-import itertools
-import functools
-
+import collections
 import dill
+import fnmatch
+import functools
+import itertools
+import multiprocessing
+import os
+import Queue
+import random
+import threading
+import time
 
 from pacer.std_logger import get_logger
 
@@ -18,10 +18,18 @@ def remote_eval(f_s, args_s):
     """ requires function and args serialized with dill, so we can remote execute even lambdas or
     decorated functions or static methods, which 'pure' multiprocssing can not handle"""
 
-    args = dill.loads(args_s)
-    f = dill.loads(f_s)
-    get_logger(f).info("start {}{} in process {}".format(f.__name__, args, os.getpid()))
     try:
+        args = dill.loads(args_s)
+    except Exception, e:
+        get_logger(remote_eval).error("got exception when unpicling args: {}".format(e))
+        return Exception(e.message)
+    try:
+        f = dill.loads(f_s)
+    except Exception, e:
+        get_logger(remote_eval).error("got exception when unpicling f: {}".format(e))
+        return Exception(e.message)
+    try:
+        get_logger(f).info("start {}{} in process {}".format(f.__name__, args, os.getpid()))
         result = f(*args)
         get_logger(f).info("got result {} from process {}".format(result, os.getpid()))
         return result
@@ -92,8 +100,7 @@ class Engine(object):
             # serialized as strings, which we deserialize in the remote process again:
             f_s = dill.dumps(f)
             args_s = dill.dumps(args)
-            Engine.pool.apply_async(remote_eval,
-                                    (f_s, args_s),
+            Engine.pool.apply_async(remote_eval, (f_s, args_s),
                                     callback=wrap_callback(new_enumeration, callback))
 
 
@@ -143,12 +150,6 @@ def files_from(folder, *patterns):
     return Source(files)
 
 
-class _ConstantSource(Source):
-
-    def __init__(self, value):
-        super(ConstantSource, self).__init__([value])
-
-
 class Computation(DataStream):
 
     def __init__(self, f, inputs):
@@ -197,8 +198,8 @@ class _UpdateManagerForJoin(object):
         # a list of collected data per stream:
         self.data = [list() * a.size() for a in input_streams]
 
-    def new_function_arguments_for(self, stream_index, value):
-        self.data[stream_index].append(value)
+    def new_function_arguments_for(self, stream_index, item):
+        self.data[stream_index].append(item)
         # cross product over all indices of available data:
         sizes = [len(v) for v in self.data]
         for perm in itertools.product(*(range(size) for size in sizes)):
@@ -223,8 +224,8 @@ class JoinComputation(Computation):
             started_computation = False
             for i, stream in enumerate(self.input_streams):
                 if not stream.empty():
-                    val = stream.get()
-                    for args in manager.new_function_arguments_for(i, val):
+                    item = stream.get()
+                    for args in manager.new_function_arguments_for(i, item):
                         Engine.run_async(self.f, args, callback=self.put)
                         started_computation = True
             if not started_computation:
@@ -242,12 +243,12 @@ class SummarizeComputation(Computation):
         return 1
 
     def _start_computation(self):
-        values = []
+        items = []
         stream = self.input_streams[0]
         for i in range(stream.size()):
-            values.append(stream.get())
-        values.sort(key=lambda item: item.number)
-        values = [v for __, v in values]
+            items.append(stream.get())
+        items.sort(key=lambda item: item.number)
+        values = [v for __, v in items]
         arg = EnumeratedItem(0, values)
         Engine.run_async(self.f, (arg,), callback=self.put)
 
@@ -270,42 +271,52 @@ class ZipComputation(Computation):
         # sizes has size 0 means: we only had inputs of size 1
         return 1
 
-    def _get_constant_inputs(self):
+    def _setup_computations(self):
+        cip = self._collect_data_from_constant_input_streams()
+
+    def _collect_data_from_constant_input_streams(self):
         constant_inputs = [None] * len(self.input_streams)
         for i, stream in enumerate(self.input_streams):
             if stream.size() == 1:
                 constant_inputs[i] = stream.get()
         return constant_inputs
 
-    def _check_for_new_data(self):
+    def _check_for_any_new_data(self):
         for stream in self.input_streams:
             if stream.size() > 1:
-                if stream.empty():
-                    return False
-        return True
+                if not stream.empty():
+                    return True
+        return False
 
-    def _assemble_function_arguments(self, constant_inputs):
-        args = [None] * len(self.input_streams)
+    def _assemble_function_arguments(self, collected_input_data, cip):
         for i, stream in enumerate(self.input_streams):
-            if stream.size() == 1:
-                args[i] = constant_inputs[i]
-            else:
-                args[i] = stream.get()
-        return tuple(args)
+            if not stream.empty():
+                item = stream.get()
+                if item.number not in collected_input_data.keys():
+                    collected_input_data[item.number] = cip
+                collected_input_data[item.number][i] = item
+        for key, arg in collected_input_data.items():
+            if all(c is not None for c in arg):
+                yield arg
+            del collected_input_data[key]
 
     def _start_computation(self):
         n = 0
         # wait and get the constant args which we need anyway:
-        constant_inputs = self._get_constant_inputs()
-        while n < self.size():
-            avail = self._check_for_new_data()
-            if avail:
-                args = self._assemble_function_arguments(constant_inputs)
-                Engine.run_async(self.f, args, callback=self.put)
-                n += 1
-            else:
-                time.sleep(0.001)
-
+        cip = self._collect_data_from_constant_input_streams()
+        collected_input_data = dict()
+        # only constant inputs ?
+        if self.size() == 1:
+            Engine.run_async(self.f, cip, callback=self.put)
+        else:
+            while n < self.size():
+                avail = self._check_for_any_new_data()
+                if avail:
+                    for args in self._assemble_function_arguments(collected_input_data, cip):
+                        Engine.run_async(self.f, args, callback=self.put)
+                        n += 1
+                else:
+                    time.sleep(0.001)
 
 def apply(inner):
     @functools.wraps(inner)
